program xxref; { MASM multifile cross-reference utility }
{ Copyright 1987 by Chrysalis Software Corp. All rights reserved.

  from: DDJ oct 1988 "A double cross for MASM"

  Last Update : Oct 23,1988 by J R Ferguson
                - adaption to Turbo Pascal 4.0
                - optimization of time-critical routines

  Turbo Pascal v3.0/4.0 MS-DOS
}

{ if Turbo Pascal v3.0 }
(*
  type
    word      = integer;
    registers = record case integer of
                  0: (ax,bx,cx,dx,bp,si,di,ds,es,flags: word);
                  1: (al,ah,bl,bh,cl,ch,dl,dh         : byte);
                end;
*)
{ endif }


{ if Turbo Pascal v4.0 }
  uses {[  RTSlib,  ]} Dos;
{ endif }

{$I-}
{$V-,R-}

const
  PubSize                      = 10000;
  RefSize                      = 4096;
  RefShift                     = 12;
  LegalChars     : set of char = ['A'..'Z','0'..'9','@','_'];
  Blanks         : string[33]  = '                                 ';
  Separator                    = '.................................';

type
  AnyString      = string[255];
  SomeString     = string[32];
  StrPtrArr      = array[1..PubSize] of ^SomeString;
  SortArray      = array[char] of integer;
  NameArray      = array[1..255] of ^SomeString;
  FileNumArray   = array[1..PubSize] of byte;
  RefPtr         = record First, Last: integer end;
  Ref            = record
                     FileNum  : byte;
                     Define   : boolean;
                     LineNum  : integer;
                     Next     : integer;
                   end;
  RefPtrArr      = array[1..PubSize] of RefPtr;
  RefArr         = array[1..RefSize] of Ref;

var
  PublicVarArr   : ^StrPtrArr;
  PublicFileArr  : FileNumArray;
  Junk           : AnyString;
  Line           : AnyString;
  Temp           : SomeString;
  i,j,k,l,m,n    : integer;
  RAnum          : integer;
  RAoff          : integer;
  PublicCount    : integer;
  ExtrnCount     : integer;
  InFile         : text;
  InFileName     : AnyString;
  OutFile        : text;
  OutFileName    : AnyString;
  KeyLen         : integer;
  ArrayLength    : integer;
  FileName       : NameArray;
  FileCount      : integer;
  Name           : AnyString;
  Ext            : SomeString;
  PosSep         : integer;
  Token          : SomeString;
  TempToken      : SomeString;
  TempRef        : Ref;
  StartOfLine    : boolean;
  FoundDef       : integer;
  OutLine        : integer;
  LastSymLine    : AnyString;
  SameSym        : boolean;
  SendCRs        : boolean;

  RefPtrs        : ^RefPtrArr;
  RefArrs        : array[1..16] of ^RefArr;
  RefCount       : integer;
  Page           : integer;
  Brief          : boolean;
  TopOfHeap      : ^integer;


procedure IOcheck(var FileName: SomeString);
var
  IOcode : integer;
  Ch     : char;
begin
{[  RTSstart(1);  ]}
  IOCode:= IOresult;
  if IOcode <> 0 then begin
    WriteLn;
    case IOcode of
      $01: WriteLn('File ',FileName,' does not exist');
      $02: WriteLn('File ',FileName,' not open for input');
      $03: WriteLn('File ',FileName,' not open for output');
      $04: WriteLn('File ',FileName,' not open');
      $05: WriteLn('Can''t write to ',FileName);
      $06: WriteLn('Can''t read from ',FileName);
      $F0: WriteLn('Disk write error on ',FileName);
      $F1: WriteLn('Directory is full');
      $FF: WriteLn('File ',FileName,' disappeared');
    else   WriteLn('Unexpected I/O error on ',FileName,', IOresult = ',IOcode);
    end;
    Halt;
  end;
{[  RTSstop(1);  ]}
end;


function GetDateStr: SomeString;
const
  Dos_get_date = $2A;
  Dos_get_time = $2C;

var
  regs   : registers;
  result : SomeString;
  temp   : SomeString;
begin with regs do begin
  ah:= Dos_get_date; MsDos(regs); Str(dh,result);
  Str(dl,temp); result:= result + '/' + temp;
  Str(cx,temp); result:= result + '/' + temp;
  ah:= Dos_get_time; MsDos(regs);
  str(ch,temp); if Length(temp) = 1 then temp:= '0' + temp;
  result:= result + ' ' + temp;
  Str(cl,temp); if Length(temp) = 1 then temp:= '0' + temp;
  result:= result + ':' + temp;
  Str(dh,temp); if Length(temp) = 1 then temp:= '0' + temp;
  result:= result + ':' + temp;
  GetDateStr:= result;
end end;


procedure MegaSort(var PtrArray : StrPtrArr;
                   var SubArray1: FileNumArray;
                   var SubArray2: RefPtrArr;
                       KeyLength: integer;
                       ArraySize: integer);
var
  l,m            : char;
  i,j            : integer;
  BucketCount    : SortArray;
  BucketPosition : SortArray;
  TempPtrArr     : ^StrPtrArr;
  TempSubArr1    : FileNumArray;
  TempSubArr2    : ^RefPtrArr;
begin
{[  RTSstart(2);  ]}
  new(TempPtrArr); new(TempSubArr2);
  for i:= KeyLength downto 1 do begin
    for l:= #0 to #255 do BucketCount[l]:= 0;
    for j:= 1 to ArraySize do begin
      if i > Length(PtrArray[j]^) then m:= #0 else m:= PtrArray[j]^[i];
      BucketCount[m]:= BucketCount[m] + 1;
    end;

    BucketPosition[#0]:= 1;
    for l:= #1 to #255 do
      BucketPosition[l]:= BucketCount[pred(l)] + BucketPosition[pred(l)];

    for j:= 1 to ArraySize do begin
      if i > Length(PtrArray[j]^) then m:= #0 else m:= PtrArray[j]^[i];
      TempPtrArr^ [BucketPosition[m]]:= PtrArray [j];
      TempSubArr1 [BucketPosition[m]]:= SubArray1[j];
      TempSubArr2^[BucketPosition[m]]:= SubArray2[j];
      BucketPosition[m]:= BucketPosition[m] + 1;
    end;

    for j:= 1 to ArraySize do begin
      PtrArray [j]:= TempPtrArr^ [j];
      SubArray1[j]:= TempSubArr1 [j];
      SubArray2[j]:= TempSubArr2^[j];
    end;
  end;
{[  RTSstop(2);  ]}
end;


procedure Canonical(var old: AnyString);
var i  : byte;
    len: byte absolute old;
begin
{[  RTSstart(3);  ]}
  if len > 0 then begin
    i:= pos(';',old);
    if i > 0 then len:= i-1;
    for i:= 1 to len do begin
      case old[i] of
              #9: old[i]:= ' ';
        'a'..'z': old[i]:= UpCase(old[i]);
      end;
    end;
  end;
{[  RTSstop(3);  ]}
end;


function Hash(var Name: SomeString): integer;
var i,temp : integer;
    len    : byte absolute Name;
begin
{[  RTSstart(4);  ]}
  if len = 0 then Hash:= 0
  else begin
    temp:= 0;
    for i:= 1 to len do
      temp:= ((temp shl 5) + (temp shr 11) + ord(Name[i])) and 32767;
    Hash:= (temp mod PubSize) + 1;
  end;
{[  RTSstop(4);  ]}
end;


function Locate(var Value: SomeString): integer;
var i: integer;
begin
{[  RTSstart(5);  ]}
  i:= Hash(Value);
  while (PublicVarArr^[i] <> nil) and (PublicVarArr^[i]^ <> Value) do begin
    if i >= PubSize then i:= 1 else i:= i+1;
  end;
  if PublicVarArr^[i] = nil then Locate:= 0 else Locate:= i;
{[  RTSstop(5);  ]}
end;

procedure RemoveToken(var Line: AnyString; var Token: SomeString);
var i,j    : byte;
    LinLen : byte absolute Line;
begin
{[  RTSstart(6);  ]}
  Token:= '';
  if LinLen <> 0 then begin
    i:= 1;
    while (i <= LinLen) and not (Line[i] in LegalChars) do i:= i + 1;
    if i > LinLen then Line:= ''
    else begin
      j:= i; while (j <= LinLen) and (Line[j] in LegalChars) do j:= j + 1;
      Token:= Copy(Line,i,j-i); Line:= Copy(Line,j,255);
    end;
  end;
{[  RTSstop(6);  ]}
end;


procedure NextToken(var Line: AnyString; var Token: SomeString);
var i,j   : byte;
    LinLen: byte absolute Line;
begin
{[  RTSstart(7);  ]}
  Token:= '';
  if LinLen <> 0 then begin
    i:= 1;
    while (i <= LinLen) and not (Line[i] in LegalChars) do i:= i + 1;
    if i <= LinLen then begin
      j:= i;
      while (j <= LinLen) and (Line[j] in LegalChars) do j:= j + 1;
      Token:= Copy(Line,i,j-i);
    end;
  end;
{[  RTSstop(7);  ]}
end;

function LookAhead(var Line: AnyString): boolean;
var
  i    : integer;
  Temp : boolean;
  Found: boolean;
begin
{[  RTSstart(8);  ]}
  Temp:= false; Found:= false; i:= 1;

  while (i < Length(Line)) and (Line[i] = ' ') do i:= i + 1;
  if Line[i] = ':' then begin
    if i = Length(Line) then begin Temp:= true; Found:= true end
    else begin
      repeat i:= i + 1 until (i = Length(Line)) or (Line[i] <> '');
      if i = Length(Line) then begin Temp:= true; Found:= true end;
    end;
  end;

  if not Found then if Copy(Line,i,3) = 'END' then begin
    Temp:= false; Found:= true
  end;

  if not Found then Temp:= (Line[i] in ['A'..'Z',':']);

  LookAhead:= Temp;
{[  RTSstop(8);  ]}
end;

function Store(var StrArray: StrPtrArr; Value: SomeString): integer;
var i: integer;
begin
{[  RTSstart(9);  ]}
  i:= Hash(Value);
  while (StrArray[i] <> nil) do begin
    i:= i + 1;
    if i > PubSize then i:= 1;
  end;
  Store:= i;
{[  RTSstop(9);  ]}
end;


procedure Header(Continued: boolean); forward;


procedure WriteOutCR;
begin
  if SendCRs then begin
    WriteLn(OutFile); IOcheck(OutFileName);
    if Brief then exit;
    OutLine:= OutLine + 1;
    if OutLine > 60 then begin
      WriteLn(OutFile); IOcheck(OutFileName);
      WriteLn(OutFile,'':20,'(* = line where symbol is defined)',chr(12));
      IOcheck(OutFileName);
      Header(SameSym);
    end;
  end;
end;


procedure WriteOut(Line: AnyString);
begin
  write(OutFile,Line); IOcheck(OutFileName);
  SendCRs:= true;
end;


procedure WriteOutLn(Line: AnyString);
begin
  SendCRs:= true;
  WriteLn(OutFile,Line); IOcheck(OutFileName);
  if Brief then exit;
  OutLine:= OutLine + 1;
  if OutLine > 60 then begin
    WriteLn(OutFile); IOcheck(OutFileName);
    WriteLn(OutFile,'':20,'(* = line where symbol is defined)',chr(12));
    IOcheck(OutFileName);
    Header(SameSym);
  end;
end;


procedure WriteOutSym(Line: AnyString);
var i: integer;
begin
  if Brief then begin
    WriteLn(OutFile,Line); IOcheck(OutFileName);
    SendCRs:= true;
    exit;
  end;

  if OutLine > 58 then begin
    for i:= OutLine + 1 to 61 do WriteLn(OutFile);
    WriteLn(OutFile); IOcheck(OutFileName);
    WriteLn(OutFile,'':20,'(* = line where symbol is defined)',chr(12));
    IOcheck(OutFileName);
    Header(false);
  end;
  LastSymLine:= Line;
  WriteLn(OutFile,Line); IOcheck(OutFileName);
  SendCRs:= true;
  OutLine:= OutLine + 1;
end;


procedure Header;
var i   : integer;
    Temp: AnyString;
begin
  OutLine:= 0;
  Str(Page,Temp); Page:= Page + 1;
  WriteOutCR; WriteOutCR;
  WriteOutLn('       XXREF V1.0 - Copyright 1987 by Chrysalis Software Corp.'+
             '       Page '+Temp);
  WriteOutCR;
  WriteOutLn('       Cross reference as of '+GetDateStr+' among files:');
  WriteOutCR;

  for i:= 1 to FileCount do begin
    WriteOut(Copy(blanks,1,13-Length(FileName[i]^))+FileName[i]^);
    if i mod 6 = 0 then WriteOutCR;
  end;
  if FileCount mod 6 <> 0 then WriteOutCR;
  WriteOutCR;
  WriteOutCR;

            { ....v....1....v....2....v....3... }
  WriteOutLn('Public symbol                    Declared in file');
  WriteOutLn('-------------                    ----------------');
  WriteOutCR;

  if Continued then begin
    LastSymLine:= Copy(LastSymLine,1,33) + '(Continued)';
    WriteOutLn(LastSymLine);
    WriteOutCR
  end;

  SendCRs:= false;
end;


procedure Initialize;
begin
  Mark(TopOfHeap);
  New(PublicVarArr); New(RefPtrs);

  if ParamCount = 0 then begin
    WriteLn('The parameters are (in this order):');
    WriteLn('Input file name ' +
            '(which contains the names of the files to process)');
    WriteLn('Output file name, where the report will be written');
    WriteLn('Optional switch: -b (produce brief output, no page headers)');
    Halt;
  end;

  if ParamCount = 1 then begin
    InFileName:= ParamStr(1);
    Write('Output file name    : '); ReadLn(OutFileName);
    Write('Brief output? (Y/N) : '); ReadLn(Temp); Temp:= UpCase(Temp[1]);
    Brief:= Temp='Y';
  end

  else if ParamCount = 2 then begin
    InFileName:= ParamStr(1);
    OutFileName:= ParamStr(2);
    Brief:= false;
  end

  else if ParamCount = 3 then begin
    InFileName:= ParamStr(1);
    OutFileName:= ParamStr(2);
    Temp:= ParamStr(3); Brief:= UpCase(Temp[2]) = 'B';
  end;

  Writeln('Reading');
  Assign (InFile, InFileName);  IOcheck(InFileName);
  Reset  (InFile);              IOcheck(InFileName);
  Assign (OutFile,OutFileName); IOcheck(OutFileName);
  Rewrite(OutFile);             IOcheck(OutFileName);

  i:= 0;
  while not eof(InFile) do begin
    ReadLn(InFile,Line); IOcheck(InFileName);
    if  (Line <> '') and (Line[1] <> ' ') and (Line[1] <> '.') and
       (Copy(Line,13,9) <> '        0')
    then begin
      i:= i + 1;
      Name:= Copy(Line,1,8); Ext:= Copy(Line,10,3);
      if pos(' ',Name) > 0 then name:= Copy(Name,1,pos(' ',Name)-1);
      Name:= Name + '.' + Ext;
      GetMem(FileName[i],Length(Name)+1);
      FileName[i]^:= Name;
    end;
  end;

  Close(InFile); IOcheck(InFileName);
  FileCount:= i;
  PublicCount:= 0;
  for i:= 1 to PubSize do PublicVarArr^[i]:= nil;
end;

procedure Pass1;

begin
{[  RTSstart(10);  ]}
  Writeln('Pass 1');

  for i:= 1 to FileCount do begin
    Write(FileName[i]^:13); if i mod 6 = 0 then WriteLn;
    Assign(InFile,FileName[i]^); IOcheck(FileName[i]^);
    Reset (InFile);              IOcheck(FileName[i]^);
    while not eof(InFile) do begin
      ReadLn(InFile,Junk); IOcheck(FileName[i]^);
      Canonical(Junk);
      NextToken(Junk,temp);
      if temp = 'PUBLIC' then begin
        RemoveToken(Junk,Temp);
        repeat
          RemoveToken(Junk,Temp);
          if Temp <> '' then begin
            j:= Locate(Temp);
            if j = 0 then begin
              j:= Store(PublicVarArr^,Temp);
              GetMem(PublicVarArr^[j],Length(Temp)+1);
              PublicVarArr^[j]^:= Temp;
              PublicFileArr[j]:= i;
              PublicCount:= PublicCount + 1;
            end;
          end;
        until Junk = '';
      end;
    end;
    Close(InFile); IOcheck(InFileName);
  end;
  WriteLn; WriteLn;
{[  RTSstop(10);  ]}
end;


procedure Pass2;
begin
{[  RTSstart(11);  ]}
  WriteLn('Pass 2');

  for i:= 1 to PubSize do RefPtrs^[i].First:= 0;
  RefCount:= 0;

  for i:= 1 to FileCount do begin
    k:= 0;
    Write(FileName[i]^:13);
    if i mod 6 = 0 then WriteLn;
    Assign(InFile,FileName[i]^); IOcheck(FileName[i]^);
    Reset (InFile);              IOcheck(FileName[i]^);
    while not eof(InFile) do begin
      ReadLn(InFile,Junk);       IOcheck(FileName[i]^);
      k:= k + 1;
      StartOfLine:= true;
      Canonical(Junk);
      RemoveToken(Junk,Token);
      if Token <> 'PUBLIC' then begin
        while Token <> '' do begin
          j:= Locate(Token);
          if j <> 0 then begin
            RefCount:= RefCount + 1;
            if RefCount mod RefSize = 1 then
              new(RefArrs[RefCount div RefSize + 1]);
            if RefPtrs^[j].First = 0 then RefPtrs^[j].First:= RefCount
            else begin
              l:= RefPtrs^[j].Last;
              RAnum:= ((l-1) shr RefShift) + 1;
              RAoff:= ((l-1) and (RefSize-1)) + 1;
              RefArrs[RAnum]^[RaOff].Next:= RefCount;
            end;
            RefPtrs^[j].Last:= RefCount;
            TempRef.FileNum:= i;
            TempRef.LineNum:= k;
            Tempref.Next:= 0;
            TempRef.Define:= false;
            if StartOfLine then TempRef.Define:= LookAhead(Junk);
            RAnum:= ((RefCount-1) shr RefShift) + 1;
            RAoff:= ((RefCount-1) and (RefSize-1)) + 1;
            RefArrs[RAnum]^[RAoff]:= TempRef;
          end;
          StartOfLine:= false;
          RemoveToken(Junk,Token);
        end;
      end;
    end;
    Close(InFile); IOcheck(FileName[i]^);
  end;
{[  RTSstop(11);  ]}
end;


procedure CompactResults;
begin
{[  RTSstart(12);  ]}
  j:= 1; k:= 1;
  for i:= 1 to PubSize do begin
    if PublicVarArr^[i] <> nil then begin
      PublicVarArr^[j]:= PublicVarArr^[i];
      PublicFileArr[j]:= PublicFileArr[i];
      RefPtrs^[j]     := RefPtrs^[i];
      if i > j then begin
        PublicVarArr^[i] := nil;
        PublicFileArr[i] := 0;
        RefPtrs^[i].First:= 0;
        RefPtrs^[i].Last := 0;
      end;
      j:= j + 1;
    end;
  end;
  WriteLn; WriteLn;
{[  RTSstop(12);  ]}
end;


procedure WriteResults;
begin
{[  RTSstart(13);  ]}
  Writeln('Writing output file');

  Page:= 1; LastSymLine:= ''; SendCRs:= true; Header(false);
  for i:= 1 to PublicCount do begin
    WriteOutCR;
    WriteOutSym(PublicVarArr^[i]^ +
                Copy(separator,1,33-Length(PublicVarArr^[i]^)) +
                FileName[PublicFileArr[i]]^);
    SameSym:= true;
    if RefPtrs^[i].First = 0 then WriteOutCR
    else begin
      k:= 0; l:= 0; j:= RefPtrs^[i].First;
      repeat
        RAnum:= ((j-1) shr RefShift) + 1;
        RAoff:= ((j-1) and (RefSize-1)) + 1;
        m:= RefArrs[RAnum]^[RAoff].FileNum;
        Junk:= FileName[RefArrs[RAnum]^[RAoff].FileNum]^;
        if k = m then begin
          if l >= 8 then begin
            WriteOutCR;
            WriteOut(Copy(Blanks,1,18));
            l:= 0;
          end;
        end
        else begin
          l:= 0;
          WriteOutCR;
          WriteOut('     ');
          WriteOut(Junk + Copy(Blanks,1,13-Length(Junk)));
        end;

        Str(RefArrs[RAnum]^[RAoff].LineNum,Temp);
        Temp:= Copy(Blanks,1,6-Length(Temp)) + Temp;
        WriteOut(Temp);
        if RefArrs[RAnum]^[RAoff].Define then WriteOut('*')
        else WriteOut(' ');
        l:= l+1;
        k:= m;
        j:= RefArrs[RAnum]^[RAoff].Next;
      until j = 0;
      SameSym:= false;
      WriteOutCR;
      WriteOutCR;
    end;
  end;

  WriteOutCR;
  Str(PublicCount,Temp); WriteOutLn('Number of publics   : ' + Temp);
  Str(RefCount   ,Temp); WriteOutLn('Number of references: ' + Temp);
  WriteOut(chr(12));
{[  RTSstop(13);  ]}
end;


procedure Terminate;
begin
  Close(OutFile); IOcheck(OutFileName);
  WriteLn('Done');
  Release(TopOfHeap);
end;

begin { main program }
{[  RTSinitAll;  ]}
{[  RTSstart(14);  ]}
  Initialize;
  Pass1;
  Pass2;
  CompactResults;
  Writeln('Sorting');
  MegaSort(PublicVarArr^,PublicFileArr,RefPtrs^,32,PublicCount);
  WriteResults;
  Terminate;
{[  RTSstop(14);  ]}
{$I+}
{[  writeln('RTS report'); RTSreport('CON');  ]}
end.
